unit AdventurePoolUnit;

interface

uses
  DosAdventure, System.Generics.Collections;

type
  TDosAdventureSync = class
  private
    FDosAdventure: TDosAdventure;
    FHasReponse: Boolean;
    FResponse: string;
    procedure InitResponse;
    procedure NotifyResponse(Sender: TObject; const AResponse: string);
    function WaitForResponse(out AResponse: string): Boolean;
  public
    constructor Create;
    destructor Destroy; override;
    function IsRunning: Boolean;
    function SendText(const AText: string; out AResponse: string): Boolean; overload;
    procedure SendText(const AText: string); overload;
    function Start(out AResponse: string): Boolean; overload;
    procedure Start; overload;
  end;

type
  TAdventurePool = class
  strict private
  private type
    TAdventureDict = TObjectDictionary<string,TDosAdventureSync>;
  private class var
    FInstance: TAdventurePool;
    class constructor Create;
    class destructor Destroy;
  private
    FAdventureDict: TAdventureDict;
  protected
    class function Instance: TAdventurePool;
    function InternalCreateAdventure(out ID: string): TDosAdventureSync;
    function InternalFindById(const ID: string): TDosAdventureSync;
    procedure InternalReleaseAll;
    procedure InternalReleaseId(const ID: string);
  public
    constructor Create;
    destructor Destroy; override;
    class procedure ReleaseAll;
    class function SendText(const ID, AText: string; out AResponse: string; out AIsRunning: Boolean): Boolean; overload;
    class function Start(out ID, AResponse: string; out AIsRunning: Boolean): Boolean;
  end;

implementation

{$DEFINE USE_CODESITE}

uses
  {$IFDEF USE_CODESITE}CodeSiteLogging,{$ENDIF}
  Winapi.Windows, System.SysUtils, System.Classes, DosAdventure.Embedded;

class constructor TAdventurePool.Create;
begin
  FInstance := TAdventurePool.Create;
  TDosAdventureFactory.Instance := TDosAdventureEmbeddedFactory.Create(false);
end;

class destructor TAdventurePool.Destroy;
begin
  FInstance.Free;
end;

constructor TAdventurePool.Create;
begin
  inherited Create;
  FAdventureDict := TAdventureDict.Create([doOwnsValues]);
end;

destructor TAdventurePool.Destroy;
begin
  FAdventureDict.Free;
  inherited Destroy;
end;

function TAdventurePool.InternalCreateAdventure(out ID: string): TDosAdventureSync;
begin
  {$IFDEF USE_CODESITE}CodeSite.TraceMethod(Self, 'InternalCreateAdventure');{$ENDIF}

  Result := TDosAdventureSync.Create;
  TMonitor.Enter(FAdventureDict);
  try
    repeat
      ID := TGUID.NewGuid.ToString;
    until not FAdventureDict.ContainsKey(ID);
    FAdventureDict.Add(ID, Result);
  finally
    TMonitor.Exit(FAdventureDict);
  end;
end;

function TAdventurePool.InternalFindById(const ID: string): TDosAdventureSync;
begin
  {$IFDEF USE_CODESITE}CodeSite.TraceMethod(Self, 'InternalFindById');{$ENDIF}

  TMonitor.Enter(FAdventureDict);
  try
    if not FAdventureDict.TryGetValue(ID, Result) then begin
      Result := nil;
    end;
  finally
    TMonitor.Exit(FAdventureDict);
  end;
end;

class function TAdventurePool.Instance: TAdventurePool;
begin
  Result := FInstance;
end;

procedure TAdventurePool.InternalReleaseAll;
begin
  {$IFDEF USE_CODESITE}CodeSite.TraceMethod(Self, 'InternalReleaseAll');{$ENDIF}
  {$IFDEF USE_CODESITE}CodeSite.Send('Dict.Count', FAdventureDict.Count);{$ENDIF}

  TMonitor.Enter(FAdventureDict);
  try
    FAdventureDict.Clear;
  finally
    TMonitor.Exit(FAdventureDict);
  end;
end;

procedure TAdventurePool.InternalReleaseId(const ID: string);
begin
  {$IFDEF USE_CODESITE}CodeSite.TraceMethod(Self, 'InternalReleaseId');{$ENDIF}

  TMonitor.Enter(FAdventureDict);
  try
    FAdventureDict.Remove(ID);
  finally
    TMonitor.Exit(FAdventureDict);
  end;
end;

class procedure TAdventurePool.ReleaseAll;
begin
  {$IFDEF USE_CODESITE}CodeSite.TraceMethod('TAdventurePool.ReleaseAll');{$ENDIF}

  Instance.InternalReleaseAll;
end;

class function TAdventurePool.SendText(const ID, AText: string; out AResponse: string; out AIsRunning: Boolean):
    Boolean;
var
  adventure: TDosAdventureSync;
begin
  {$IFDEF USE_CODESITE}CodeSite.TraceMethod('TAdventurePool.SendText');{$ENDIF}

  result := false;
  adventure := Instance.InternalFindById(ID);
  if adventure <> nil then begin
    try
      Result := adventure.SendText(AText, AResponse);
      AIsRunning := adventure.IsRunning;
    except
      on E: Exception do begin
        AIsRunning := false;
        AResponse := E.Message;
        Result := false;
      end;
    end;
    if not AIsRunning then begin
      Instance.InternalReleaseId(ID);
    end;
  end;
end;

class function TAdventurePool.Start(out ID, AResponse: string; out AIsRunning: Boolean): Boolean;
var
  adventure: TDosAdventureSync;
begin
  {$IFDEF USE_CODESITE}CodeSite.TraceMethod('TAdventurePool.Start');{$ENDIF}

  result := false;
  adventure := Instance.InternalCreateAdventure(ID);
  if adventure <> nil then begin
    try
      Result := adventure.Start(AResponse);
      AIsRunning := adventure.IsRunning;
    except
      on E: Exception do begin
        AIsRunning := False;
        AResponse := E.Message;
        Result := False;
      end;
    end;
    if not AIsRunning then begin
      Instance.InternalReleaseId(ID);
      ID := '';
    end;
  end;
end;

constructor TDosAdventureSync.Create;
begin
  inherited Create;
  FDosAdventure := TDosAdventureFactory.CreateDosAdventure;
  FDosAdventure.OnNotifyResponse := NotifyResponse;
end;

destructor TDosAdventureSync.Destroy;
begin
  FDosAdventure.Free;
  inherited Destroy;
end;

procedure TDosAdventureSync.NotifyResponse(Sender: TObject; const AResponse: string);
begin
  FResponse := AResponse;
  FHasReponse := true;
end;

function TDosAdventureSync.SendText(const AText: string; out AResponse: string): Boolean;
begin
  SendText(AText);
  Result := WaitForResponse(AResponse);
end;

function TDosAdventureSync.Start(out AResponse: string): Boolean;
begin
  Start;
  result := WaitForResponse(AResponse);
end;

procedure TDosAdventureSync.Start;
begin
  InitResponse;
  if FDosAdventure <> nil then begin
    FDosAdventure.Start;
  end;
end;

procedure TDosAdventureSync.InitResponse;
begin
  if FDosAdventure <> nil then
  begin
    FResponse := '';
    FHasReponse := False;
  end
  else
  begin
    FResponse := 'not implemented';
    FHasReponse := True;
  end;
end;

function TDosAdventureSync.IsRunning: Boolean;
begin
  Result := (FDosAdventure <> nil) and FDosAdventure.IsRunning;
end;

procedure TDosAdventureSync.SendText(const AText: string);
begin
  InitResponse;
  if FDosAdventure <> nil then begin
    FDosAdventure.SendText(AText);
  end;
end;

function TDosAdventureSync.WaitForResponse(out AResponse: string): Boolean;
begin
  while not FHasReponse do begin
//    Application.ProcessMessages;
    if (GetCurrentThreadID = MainThreadID) then begin
      CheckSynchronize(10);
    end
    else begin
      Sleep(10);
    end;
  end;
  AResponse := FResponse;
  result := FHasReponse;
end;

end.

